﻿
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using Microsoft.Win32;

/*    Solving a Sudoku by the use of recursive function calls    */
/*                        www.mosismath.com                      */

namespace Sudoku
{
    public partial class MainWindow : Window
    {
     
        public class TBasicCell
        {
            public int valueSet;
            public TextBox tb;
        };

        public class TBlock
        {
            public SolidColorBrush brush;
            public Collection<TBasicCell> cells = new Collection<TBasicCell>();

            public void Clear()
            {
                cells.Clear();
            }

            public bool Add(TBasicCell cell)
            {
                if (cells.Count < 9)
                    cells.Add(cell);
                if (cells.Count < 9)
                    return true;
                else
                    return false;
            }

            public bool Remove(TBasicCell cell)
            {
                int index = cells.IndexOf(cell);
                if (index >= 0)
                {
                    cells.RemoveAt(index);
                    return true;
                }
                else
                    return false;
            }
        }

        public class TCell : TBasicCell
        {
            public TBlock block;
        };

        TCell[,] field = new TCell[9, 9];
        TBlock[] gBlock = new TBlock[9];
        TBlock tempBlock = new TBlock();
        Settings settings = new Settings();
        BackgroundWorker calculationThread;
        bool solved = false;
        bool formatError = false;
        bool defFreeForm = false;
        bool modifyFreeForm = false;
        int actBlock = 0;

        private void InitStandard()
        {
            // init standard 3*3 blocks
            int x, y, dx, dy;

            border1.BorderBrush = new SolidColorBrush(Colors.Black);
            border2.BorderBrush = border1.BorderBrush;
            border3.BorderBrush = border1.BorderBrush;
            border4.BorderBrush = border1.BorderBrush;
            border5.BorderBrush = border1.BorderBrush;
            border6.BorderBrush = border1.BorderBrush;
            border7.BorderBrush = border1.BorderBrush;
            border8.BorderBrush = border1.BorderBrush;
            border9.BorderBrush = border1.BorderBrush;
            tbActBlock.Visibility = Visibility.Hidden;
            lbActBlock.Visibility = Visibility.Hidden;
            for (x = 0; x < 9; x++)
            {
                gBlock[x].Clear();
            }
            // assign the corresponding elements to each block
            for (y = 0; y < 3; y++)
            {
                for (x = 0; x < 3; x++)
                {
                    for (dy = 0; dy < 3; dy++)
                    {
                        for (dx = 0; dx < 3; dx++)
                        {
                            gBlock[x + y * 3].Add(field[x * 3 + dx, y * 3 + dy]);
                        }
                    }
                }
            }
            // assign the corresponding block to each element
            for (y = 0; y < 3; y++)
            {
                for (x = 0; x < 3; x++)
                {
                    field[x, y].block = gBlock[0];
                    field[x + 3, y].block = gBlock[1];
                    field[x + 6, y].block = gBlock[2];
                    field[x, y + 3].block = gBlock[3];
                    field[x + 3, y + 3].block = gBlock[4];
                    field[x + 6, y + 3].block = gBlock[5];
                    field[x, y + 6].block = gBlock[6];
                    field[x + 3, y + 6].block = gBlock[7];
                    field[x + 6, y + 6].block = gBlock[8];
                }
            }
            for (x = 0; x < 9; x++)
            {
                for (y = 0; y < 9; y++)
                {
                    field[x, y].tb.Background = new SolidColorBrush(Colors.White);
                }
            }
        }

        private void InitFreeForm()
        {
            int i, j;
            for (i = 0; i < 9; i++)
            {
                gBlock[i].Clear();
            }
            tbActBlock.Visibility = Visibility.Visible;
            lbActBlock.Visibility = Visibility.Visible;
            border1.BorderBrush = new SolidColorBrush(Colors.White);
            border2.BorderBrush = border1.BorderBrush;
            border3.BorderBrush = border1.BorderBrush;
            border4.BorderBrush = border1.BorderBrush;
            border5.BorderBrush = border1.BorderBrush;
            border6.BorderBrush = border1.BorderBrush;
            border7.BorderBrush = border1.BorderBrush;
            border8.BorderBrush = border1.BorderBrush;
            border9.BorderBrush = border1.BorderBrush;
            for (i = 0; i < 9; i++)
            {
                for (j = 0; j < 9; j++)
                {
                    field[i, j].tb.Background = new SolidColorBrush(Colors.White);
                }
            }
        }

        public MainWindow()
        {
            InitializeComponent();

            calculationThread = new BackgroundWorker();
            calculationThread.DoWork += SolveSudoku;
            calculationThread.RunWorkerCompleted += SolvingCompleted;

            int x, y, i, j;
            // init the 9 blocks containing 9 cells
            for (x = 0; x < 9; x++)
            {
                gBlock[x] = new TBlock();
            }
            gBlock[0].brush = new SolidColorBrush(Colors.Lavender);
            gBlock[1].brush = new SolidColorBrush(Colors.LightBlue);
            gBlock[2].brush = new SolidColorBrush(Colors.PaleTurquoise);
            gBlock[3].brush = new SolidColorBrush(Colors.LightGreen);
            gBlock[4].brush = new SolidColorBrush(Colors.Yellow);
            gBlock[5].brush = new SolidColorBrush(Colors.LightSalmon);
            gBlock[6].brush = new SolidColorBrush(Colors.Wheat);
            gBlock[7].brush = new SolidColorBrush(Colors.LightPink);
            gBlock[8].brush = new SolidColorBrush(Colors.Plum);
            // init the 9 x 9 elements
            for (y = 0; y < 9; y++)
            {
                for (x = 0; x < 9; x++)
                {
                    field[x, y] = new TCell();
                    field[x, y].valueSet = -1;
                }
            }
                        
            // assign the text box to each element
            field[0, 0].tb = tb11; field[1, 0].tb = tb21; field[2, 0].tb = tb31;
            field[3, 0].tb = tb41; field[4, 0].tb = tb51; field[5, 0].tb = tb61;
            field[6, 0].tb = tb71; field[7, 0].tb = tb81; field[8, 0].tb = tb91;

            field[0, 1].tb = tb12; field[1, 1].tb = tb22; field[2, 1].tb = tb32;
            field[3, 1].tb = tb42; field[4, 1].tb = tb52; field[5, 1].tb = tb62;
            field[6, 1].tb = tb72; field[7, 1].tb = tb82; field[8, 1].tb = tb92;

            field[0, 2].tb = tb13; field[1, 2].tb = tb23; field[2, 2].tb = tb33;
            field[3, 2].tb = tb43; field[4, 2].tb = tb53; field[5, 2].tb = tb63;
            field[6, 2].tb = tb73; field[7, 2].tb = tb83; field[8, 2].tb = tb93;

            field[0, 3].tb = tb14; field[1, 3].tb = tb24; field[2, 3].tb = tb34;
            field[3, 3].tb = tb44; field[4, 3].tb = tb54; field[5, 3].tb = tb64;
            field[6, 3].tb = tb74; field[7, 3].tb = tb84; field[8, 3].tb = tb94;

            field[0, 4].tb = tb15; field[1, 4].tb = tb25; field[2, 4].tb = tb35;
            field[3, 4].tb = tb45; field[4, 4].tb = tb55; field[5, 4].tb = tb65;
            field[6, 4].tb = tb75; field[7, 4].tb = tb85; field[8, 4].tb = tb95;

            field[0, 5].tb = tb16; field[1, 5].tb = tb26; field[2, 5].tb = tb36;
            field[3, 5].tb = tb46; field[4, 5].tb = tb56; field[5, 5].tb = tb66;
            field[6, 5].tb = tb76; field[7, 5].tb = tb86; field[8, 5].tb = tb96;

            field[0, 6].tb = tb17; field[1, 6].tb = tb27; field[2, 6].tb = tb37;
            field[3, 6].tb = tb47; field[4, 6].tb = tb57; field[5, 6].tb = tb67;
            field[6, 6].tb = tb77; field[7, 6].tb = tb87; field[8, 6].tb = tb97;

            field[0, 7].tb = tb18; field[1, 7].tb = tb28; field[2, 7].tb = tb38;
            field[3, 7].tb = tb48; field[4, 7].tb = tb58; field[5, 7].tb = tb68;
            field[6, 7].tb = tb78; field[7, 7].tb = tb88; field[8, 7].tb = tb98;

            field[0, 8].tb = tb19; field[1, 8].tb = tb29; field[2, 8].tb = tb39;
            field[3, 8].tb = tb49; field[4, 8].tb = tb59; field[5, 8].tb = tb69;
            field[6, 8].tb = tb79; field[7, 8].tb = tb89; field[8, 8].tb = tb99;
            //initialize the elements according the the preset test boxes
            for (i = 0; i < 9; i++)
            {
                for (j = 0; j < 9; j++)
                {
                    if (field[i, j].tb.Text != "")
                    {
                        field[i, j].valueSet = Convert.ToInt32(field[i, j].tb.Text);
                        field[i, j].tb.Foreground = new SolidColorBrush(Colors.Blue);
                    }
                    else
                    {
                        field[i, j].valueSet = -1;
                        field[i, j].tb.Foreground = new SolidColorBrush(Colors.Black);
                    }
                }
            }
            InitStandard();
        }

        private bool CheckPossible()
        {
            int i, j, k;
            bool result = true;
            bool[] foundValBlock = new bool[10] { false, false, false, false, false, false, false, false, false, false };
            bool[] foundValRow = new bool[10] { false, false, false, false, false, false, false, false, false, false };
            bool[] foundValCol = new bool[10] { false, false, false, false, false, false, false, false, false, false };
            for (i = 0; i < 9; i++)
            {

                for (k = 0; k < 10; k++)
                    foundValBlock[k] = false;
                for (k = 0; k < gBlock[i].cells.Count; k++)
                {
                    if (gBlock[i].cells[k].valueSet > 0)
                    {
                        if (foundValBlock[gBlock[i].cells[k].valueSet])
                            result = false;
                        else
                            foundValBlock[gBlock[i].cells[k].valueSet] = true;
                    }
                }

            }

            for (i = 0; i < 9; i++)  // check for set values in active col
            {
                for (k = 0; k < 10; k++)
                    foundValCol[k] = false;
                for (j = 0; j < 9; j++)
                {
                    if (field[j, i].valueSet > 0)
                    {
                        if (field[j, i].valueSet <= 9)
                        {
                            if (foundValCol[field[j, i].valueSet])
                            {
                                result = false;
                            }
                            else
                            {
                                foundValCol[field[j, i].valueSet] = true;
                            }
                        }
                    }
                }
            }

            for (i = 0; i < 9; i++)  // check for set values in active col
            {
                for (k = 0; k < 10; k++)
                    foundValRow[k] = false;
                for (j = 0; j < 9; j++)
                {
                    if (field[i, j].valueSet > 0)
                    {
                        if (field[i, j].valueSet <= 9)
                        {
                            if (foundValRow[field[i, j].valueSet])
                            {
                                result = false;
                            }
                            else
                            {
                                foundValRow[field[i, j].valueSet] = true;
                            }
                        }
                    }
                }
            }
            return result;
        }


        private List<int> FindPossible(int actX, int actY)
        {
            int i, k;
            List<int> possibleList = new List<int>();
            TBlock actBlock;
            if (!(field[actX, actY].valueSet > 0))
            {
                actBlock = field[actX, actY].block;
                bool[] impossible = new bool[10] { false, false, false, false, false, false, false, false, false, false };
                for (i = 0; i < actBlock.cells.Count; i++)  // check for set values in active block
                {
                    if (actBlock.cells[i].valueSet > 0)
                    {
                        if (actBlock.cells[i].valueSet <= 9)
                        {
                            k = actBlock.cells[i].valueSet;
                            impossible[k] = true;
                        }
                    }
                }
                for (i = 0; i < 9; i++)  // check for set values in active row
                {
                    if (field[actX, i].valueSet > 0)
                    {
                        if (field[actX, i].valueSet <= 9)
                            impossible[field[actX, i].valueSet] = true;
                    }
                }
                for (i = 0; i < 9; i++)  // check for set values in active col
                {
                    if (field[i, actY].valueSet > 0)
                    {
                        if (field[i, actY].valueSet <= 9)
                            impossible[field[i, actY].valueSet] = true;
                    }
                }
                // now we have all impossible values for the actual Cell  
                if (field[actX, actY].valueSet < 0)
                {
                    for (i = 1; i <= 9; i++)
                    {
                        if (!impossible[i])
                            possibleList.Add(i);
                    }
                }
            }
            return possibleList;
        }


        private bool FillCell(int actIndex)
        {
            int i;
            bool result = true;
            bool foundOne = false;
            int nextX = -1;
            int nextY = -1;

            if (actIndex < 81)
            {
                while ((!foundOne) && (actIndex < 80))
                {
                    actIndex++;
                    nextX = actIndex % 9;
                    nextY = actIndex / 9;
                    foundOne = (field[nextX, nextY].valueSet == -1);
                }
                foundOne = false;
                if ((nextX >= 0) && (nextY >= 0))
                {
                    if (field[nextX, nextY].valueSet == -1)
                    {
                        List<int> list = FindPossible(nextX, nextY);
                        if (list.Count > 0)
                        {
                            for (i = 0; i < list.Count; i++)
                            {
                               
                                field[nextX, nextY].valueSet = list.ElementAt(i);
                                if (FillCell(actIndex))
                                {
                                    i = list.Count;
                                    foundOne = true;
                                }
                            }
                            if (!foundOne)
                            {
                                field[nextX, nextY].valueSet = -1;
                                result = false;
                            }
                        }
                        else
                        {
                            result = false;
                        }
                    }
                    else
                    {
                        result = true;
                    }
                }
            }
            else
                result = true;
            return result;
        }


        private void SolveSudoku(object sender, DoWorkEventArgs e)
        {
            formatError = !CheckPossible();
            if (!formatError)
            {
                solved = false;
                solved = FillCell(-1);
            }
        }


        private void SolvingCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            int i, j;
            if (solved && !formatError)
            {
                for (i = 0; i < 9; i++)
                {
                    for (j = 0; j < 9; j++)
                    {
                        field[i, j].tb.Text = Convert.ToString(field[i, j].valueSet);
                    }
                }
            }
            else
            {
                if (!formatError)
                    MessageBox.Show("Some thing went wrong :-(", "Solving error!");
                else
                    MessageBox.Show("Twice the same number found :-(", "Solving error!");
            }
            ctClearAll.IsEnabled = true;
            btClearResult.IsEnabled = true;
            btSolve.IsEnabled = true;
            Mouse.OverrideCursor = null;
        }

        private void Button_Click(object sender, RoutedEventArgs e)
        {
            calculationThread.RunWorkerAsync();
        }

        private void BtSolve_Click(object sender, RoutedEventArgs e)
        {
            int i;
            bool definitionOk = true;
            for (i = 0; i < 9; i++)
            {
                if (gBlock[i].cells.Count < 9)
                    definitionOk = false;
            }
            if (definitionOk)
            {
                ctClearAll.IsEnabled = false;
                btClearResult.IsEnabled = false;
                btSolve.IsEnabled = false;
                Mouse.OverrideCursor = Cursors.Wait;
                calculationThread.RunWorkerAsync();
            }
            else
            {
                MessageBox.Show("Setting of blocks not complete!");
            }
        }

        private void CtClearAll_Click(object sender, RoutedEventArgs e)
        {
            int i, j;
            for (i = 0; i < 9; i++)
            {
                for (j = 0; j < 9; j++)
                {
                    field[i, j].tb.Text = "";
                    field[i, j].valueSet = -1;
                    field[i, j].tb.Foreground = new SolidColorBrush(Colors.Black);
                }
            }
        }

        private void BtClearResult_Click(object sender, RoutedEventArgs e)
        {
            int i, j;
            for (i = 0; i < 9; i++)
            {
                for (j = 0; j < 9; j++)
                {
                    if (field[i, j].tb.Foreground.ToString() != "#FF0000FF")
                    {
                        field[i, j].tb.Text = "";
                        field[i, j].valueSet = -1;
                        field[i, j].tb.Foreground = new SolidColorBrush(Colors.Black);
                    }
                }
            }
        }

        private void Tb_KeyUp(object sender, KeyEventArgs e)
        {
            int i, j;
            int num;
            for (i = 0; i < 9; i++)
            {
                for (j = 0; j < 9; j++)
                {
                    if (field[i, j].tb.Text != "")
                    {
                        try
                        {
                            num = Convert.ToInt32(field[i, j].tb.Text);
                            if ((num < 10) && (num > 0))
                            {
                                field[i, j].valueSet = Convert.ToInt32(field[i, j].tb.Text);
                                field[i, j].tb.Foreground = new SolidColorBrush(Colors.Blue);
                            }
                            else
                            {
                                field[i, j].valueSet = -1;
                                field[i, j].tb.Foreground = new SolidColorBrush(Colors.Black);
                                field[i, j].tb.Text = "";
                            }
                        }
                        catch
                        {
                            field[i, j].valueSet = -1;
                            field[i, j].tb.Foreground = new SolidColorBrush(Colors.Black);
                            field[i, j].tb.Text = "";
                        }
                    }
                    else
                    {
                        field[i, j].valueSet = -1;
                        field[i, j].tb.Foreground = new SolidColorBrush(Colors.Black);
                        field[i, j].tb.Text = "";
                    }
                }
            }
        }

       
        private void Tb_PreviewMouseRightButtonDown(object sender, MouseButtonEventArgs e)
        {
            int i, j;
            if (sender is TextBox)
            {
                if (modifyFreeForm || defFreeForm)
                {
                    for (i = 0; i < 9; i++)
                    {
                        for (j = 0; j < 9; j++)
                        {
                            if (sender == field[i, j].tb)
                            {
                                field[i, j].tb.Background = new SolidColorBrush(Colors.White);
                                if (field[i, j].block.Remove(field[i, j]))
                                    field[i, j].block = null;
                            }
                        }
                    }
                }
            }
        }

       
        private void Tb_PreviewMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            if (sender is TextBox)
            {
                int i, j, k;
                if (modifyFreeForm || defFreeForm)
                {

                    for (i = 0; i < 9; i++)
                    {
                        for (j = 0; j < 9; j++)
                        {
                            if (sender == field[i, j].tb)
                            {
                                if (field[i, j].block == null)
                                {
                                    if (((actBlock < 8)||(modifyFreeForm && (actBlock <= 8))) && (actBlock >= 0))
                                    {
                                        if (gBlock[actBlock].cells.Count < 9)
                                        {
                                            field[i, j].tb.Background = gBlock[actBlock].brush;
                                            field[i, j].block = gBlock[actBlock];
                                            if (!gBlock[actBlock].Add(field[i, j]))
                                            {
                                                if (!modifyFreeForm)
                                                {
                                                    actBlock++;
                                                    tbActBlock.Background = gBlock[actBlock].brush;
                                                }
                                            }
                                        }
                                        else
                                            MessageBox.Show("Block is complete allready!");
                                    }
                                }
                                else
                                {
                                    for (k = 0; k < 9; k++)
                                    {
                                        if (field[i, j].block == gBlock[k])
                                        {
                                            actBlock = k;
                                            tbActBlock.Background = gBlock[actBlock].brush;
                                        }
                                    }
                                }
                            }
                        }

                    }
                    if ((actBlock == 8) && (!modifyFreeForm)) // fill the last block automatically
                        {
                        for (i = 0; i < 9; i++)
                        {
                            for (j = 0; j < 9; j++)
                            {
                                if (field[i, j].tb.Background.ToString() == "#FFFFFFFF")
                                {
                                    if (field[i, j].block == null)
                                    {
                                        if (gBlock[actBlock].cells.Count < 9)
                                        {
                                            gBlock[actBlock].Add(field[i, j]);
                                            field[i, j].tb.Background = gBlock[actBlock].brush;
                                            field[i, j].block = gBlock[actBlock];
                                            mnFreeForm.IsChecked = false;
                                        }
                                    }
                                }
                            }
                        }
                        tbActBlock.Visibility = Visibility.Hidden;
                        lbActBlock.Visibility = Visibility.Hidden;
                    }
                }
            }
        }


        private void BtStandard_Click(object sender, RoutedEventArgs e)
        {
            int i, j;

            defFreeForm = false;
            modifyFreeForm = false;
            mnStandardForm.IsChecked = true;
            mnFreeForm.IsChecked = false;
            mnModifyFree.IsEnabled = false;
            InitStandard();
            for (i = 0; i < 9; i++)
            {
                for (j = 0; j < 9; j++)
                {
                    field[i, j].tb.Background = new SolidColorBrush(Colors.White);
                }
            }            
        }

        private void BtFree_Click(object sender, RoutedEventArgs e)
        {
            int i, j;
            if (!defFreeForm)
            {
                mnStandardForm.IsChecked = false;
                mnFreeForm.IsChecked = true;
                modifyFreeForm = false;
                mnModifyFree.IsEnabled = true;
                actBlock = 0;
                defFreeForm = true;
                InitFreeForm();
                for (i = 0; i < 9; i++)
                {
                    for (j = 0; j < 9; j++)
                    {
                        field[i, j].block = null;
                    }
                }
                tbActBlock.Background = gBlock[actBlock].brush;
            }
        }

        private void btModifyClick(object sender, RoutedEventArgs e)
        {
            modifyFreeForm = !modifyFreeForm;
            mnModifyFree.IsChecked = modifyFreeForm;
            lbActBlock.Visibility = Visibility.Visible;
            tbActBlock.Visibility = Visibility.Visible;
            if (modifyFreeForm)
            {
                actBlock = -1;
                tbActBlock.Visibility = Visibility.Visible;
                lbActBlock.Visibility = Visibility.Visible;
                defFreeForm = false;
                mnStandardForm.IsChecked = false;
                mnFreeForm.IsChecked = false;
            }
            else
            {
                tbActBlock.Visibility = Visibility.Hidden;
                lbActBlock.Visibility = Visibility.Hidden;
            }
        }


        private void BtLoadSetting_Click(object sender, RoutedEventArgs e)
        {
            int actBlock, actBlockCell, x, y;
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "Settings (*json)|*.json";
            ofd.Title = "Load free form";
            ofd.InitialDirectory = System.AppDomain.CurrentDomain.BaseDirectory;

            if (ofd.ShowDialog() == true)
            {
                if (settings.readJsonData(ofd.FileName))
                {
                    InitFreeForm();
                    for (actBlock = 0; actBlock < 9; actBlock++)
                    {
                        for (actBlockCell = 0; actBlockCell < 9; actBlockCell++)
                        {
                            x = settings.parameter.block[actBlock].field[actBlockCell].x;
                            y = settings.parameter.block[actBlock].field[actBlockCell].y;
                            gBlock[actBlock].Add(field[x, y]);
                            field[x, y].block = gBlock[actBlock];
                            field[x, y].tb.Background = gBlock[actBlock].brush;
                        }
                    }
                }
                defFreeForm = false;
                modifyFreeForm = false;
                mnStandardForm.IsChecked = false;
                mnFreeForm.IsChecked = false;
                mnModifyFree.IsChecked = false;
                mnModifyFree.IsEnabled = true;
                tbActBlock.Visibility = Visibility.Hidden;
                lbActBlock.Visibility = Visibility.Hidden;
            }
        }

        private void BtSaveSetting_Click(object sender, RoutedEventArgs e)
        {
            int actBlock, actBlockCell, x, y;
            bool ErrorNotComplete = false;
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "Settings (*json)|*.json";
            sfd.Title = "Load free form";
            sfd.InitialDirectory = System.AppDomain.CurrentDomain.BaseDirectory;

            if (sfd.ShowDialog() == true)
            {
                for (actBlock = 0; actBlock < 9; actBlock++)
                {
                    if (gBlock[actBlock].cells.Count == 9)
                    {
                        for (actBlockCell = 0; actBlockCell < gBlock[actBlock].cells.Count; actBlockCell++)
                        {
                            for (x = 0; x < 9; x++)
                            {
                                for (y = 0; y < 9; y++)
                                {
                                    if (gBlock[actBlock].cells[actBlockCell] == field[x, y])
                                    {
                                        settings.parameter.block[actBlock].field[actBlockCell].x = x;
                                        settings.parameter.block[actBlock].field[actBlockCell].y = y;
                                    }
                                }
                            }
                        }
                    }
                    else
                        ErrorNotComplete = true;
                }
                if (!ErrorNotComplete)
                    settings.writeJsonData(sfd.FileName);
                else
                    MessageBox.Show("Block definitions not complete!");
            }
        }
    }
}
